@@ -32,6 +32,7 @@
 #include <unistd.h>
 #include <stdlib.h>
 #include <sys/socket.h>
+#include <netinet/in.h>
 #include <libbonobo.h>
 #include <netdb.h>
 #include "festivalsynthesisdriver.h"
@@ -44,6 +45,8 @@
 #undef FESTIVAL_DEBUG_SEND
 #undef FESTIVAL_DEBUG_TEXT
 
+#define DEFAULT_ENCODING "ISO-8859-1"
+
 static gint 		text_id   = 0;
 static GObjectClass 	*parent_class;
 static gboolean 	festival_server_exists = FALSE;
@@ -51,6 +54,7 @@ static gboolean 	festival_server_got_voices = FALSE;
 static GSList		*driver_list = NULL;
 static GSList		*markers_list = NULL;
 static GSList		*voices_list = NULL;
+static GSList		*encodings_list = NULL;
 
 typedef struct
 {
@@ -97,6 +101,37 @@ static void     festival_synthesis_driver_process_list 		(FestivalSynthesisDrive
 static void 	festival_process_text_out 			(FestivalTextOut *text_out);
 static void 	festival_free_list 				(FestivalSynthesisDriver *d);
 
+static void encoding_set(GNOME_Speech_VoiceInfo* voice, gchar* encoding)
+{
+	/* First, try to see if we have it already */
+	GSList* cur = encodings_list;
+	int len = strlen(voice->name);
+	for ( ; cur != NULL; cur = cur->next)
+		if (strncmp((gchar*)cur->data, voice->name, len) == 0)
+		{
+			/* If we have it, replace it */
+			g_free(cur->data);
+			cur->data = g_strdup_printf("%s:%s", voice->name, encoding);
+			return;
+		}
+	/* Otherwise, insert it */
+	encodings_list = g_slist_prepend (encodings_list, g_strdup_printf("%s:%s", voice->name, encoding));
+}
+
+static gchar* encoding_get(FestivalSpeaker* speaker)
+{
+	/* Look for the item in the list */
+	GSList* cur = encodings_list;
+	int len = strlen(speaker->voice);
+	/* Speaker should be at least "(voice_?)\n" */
+	if (len < 10)
+		return DEFAULT_ENCODING;
+	for ( ; cur != NULL; cur = cur->next)
+		if (strncmp((gchar*)cur->data, speaker->voice + 7, len - 9) == 0)
+			return (gchar*)cur->data + len - 9 + 1;
+	/* If not found, default to latin1 */
+	return DEFAULT_ENCODING;
+}
 
 static FestivalTextMarker *
 festival_text_marker_new ()
@@ -312,7 +347,6 @@ festival_response_sock (GIOChannel   *source,
 #ifdef FESTIVAL_DEBUG_CONTROL
     	    fprintf (stderr, "\nRECEIVE QUERY");
 #endif
-	    driver->is_querying = FALSE;
         }
 	else if (ack && strncmp (ack, "(VOICES:", 8) == 0)
 	{
@@ -329,12 +363,19 @@ festival_response_sock (GIOChannel   *source,
 		p = strstr (voices[i], ")\n");
 		if (p != NULL)
 		    *p = 0; 
-		query_voice_des = g_strdup_printf ("(list 'VOICEDESC: "
-						   "(nth 0 (voice.description '%s)) "
-						   "(nth 1 (nth 0 (nth 1 (voice.description '%s)))) "
-						   "(nth 1 (nth 2 (nth 1 (voice.description '%s)))) "
-						   "(nth 1 (nth 1 (nth 1 (voice.description '%s)))))\n",
-						   voices[i], voices[i], voices[i], voices[i]);
+		/* Adds an extra nil to help tokenization later */
+		query_voice_des = g_strdup_printf (
+			"(list 'VOICEDESC: "
+			"(nth 0 (voice.description '%s)) "
+			"(cadr (assoc 'language (cadr (voice.description '%s)))) "
+			"(cadr (assoc 'dialect  (cadr (voice.description '%s)))) "
+			"(cadr (assoc 'gender   (cadr (voice.description '%s)))) "
+			"(or "
+			"  (cadr (assoc 'coding (cadr (voice.description '%s)))) "
+			"  \"" DEFAULT_ENCODING "\") "
+			"nil)\n",
+				voices[i], voices[i], voices[i], voices[i], voices[i]);
+
 		festival_synthesis_driver_say_raw (driver, query_voice_des);
 		voices_waiting_for_description++;
 		g_free (query_voice_des);
@@ -349,7 +390,8 @@ festival_response_sock (GIOChannel   *source,
 		voice[1] != NULL && strcmp (voice[1], "nil") && 
 		voice[2] != NULL && strcmp (voice[2], "nil") && 
 		voice[3] != NULL && strcmp (voice[3], "nil") && 
-		voice[4] != NULL && strcmp (voice[4], "nil"))
+		voice[4] != NULL && strcmp (voice[4], "nil") &&
+		voice[5] != NULL && strcmp (voice[5], "nil"))
 	    {
 		GNOME_Speech_VoiceInfo *new_info;
 		new_info = GNOME_Speech_VoiceInfo__alloc ();
@@ -360,6 +402,7 @@ festival_response_sock (GIOChannel   *source,
 		else
 		    new_info->gender = GNOME_Speech_gender_male;
 		voices_list = g_slist_append (voices_list, new_info);
+		encoding_set(new_info, voice[5]);
 	    }
 	    voices_waiting_for_description--;
 	    if (voices_waiting_for_description == 0) 
@@ -482,11 +525,11 @@ festival_start (FestivalSynthesisDriver *d)
 	    return FALSE;
 	
         d->channel_pipe = g_io_channel_unix_new(d->pipe);
-	g_io_channel_set_encoding (d->channel_pipe, "ISO-8859-1", NULL);
+	g_io_channel_set_encoding (d->channel_pipe, "UTF-8", NULL);
         g_io_add_watch (d->channel_pipe, G_IO_IN , festival_response_pipe, d);
 	
 	d->channel_sock = g_io_channel_unix_new(d->sock);
-	g_io_channel_set_encoding (d->channel_sock, "ISO-8859-1", NULL);
+	g_io_channel_set_encoding (d->channel_sock, "UTF-8", NULL);
         g_io_add_watch (d->channel_sock, G_IO_IN, festival_response_sock, d);
 
 	if (festival_server_exists) {
@@ -792,7 +835,6 @@ festival_synthesis_driver_init (FestivalSynthesisDriver *d)
         d->crt_clbs = NULL;
         d->is_shutting_up = FALSE;
 	d->is_speaking = FALSE;
-        d->is_querying = FALSE;
         d->list = NULL;
         d->crt_id = 0;
         d->queue_length = 0;	
@@ -847,15 +889,37 @@ void
 festival_synthesis_driver_say_raw (FestivalSynthesisDriver *d,
 			    	   gchar 		   *text)
 {
-        int l, written;
+	GIOStatus status;
+	GError *err = NULL;
+        gsize l, written;
+
+	if (!festival_server_exists) {
+	        return;
+        }
 
 	g_assert (text && IS_FESTIVAL_SYNTHESIS_DRIVER (d) && d->channel_sock);
 #ifdef FESTIVAL_DEBUG_SEND
         fprintf (stderr, "\nRAW SEND :\"%s\"", text);
 #endif
 	l = strlen (text);
-        g_io_channel_write_chars (d->channel_sock, text, l, &written, NULL);
+        status = g_io_channel_write_chars (d->channel_sock, text, l, &written, &err);
         g_io_channel_flush (d->channel_sock, NULL);
+	if (status == G_IO_STATUS_ERROR)
+	{
+	        if (err->domain == G_CONVERT_ERROR) {
+	                fprintf (stderr,
+			         "STRING CONVERSION FAILED: (string=%s) (error=%s)!\n",
+			         text, 
+				 err->message);
+	        } else {
+	                kill (d->pid, 9);
+ 	                fprintf (stderr, 
+		                 "g_io_channel_write_chars gave G_IO_STATUS_ERROR: %s\n",
+		                 err->message);
+	                fprintf (stderr, "ABORTING.\n");
+	                exit (1);
+		}
+	}
 }
 
 
@@ -886,14 +950,10 @@ festival_querying_queue (gpointer data)
 
 	g_assert (IS_FESTIVAL_SYNTHESIS_DRIVER (driver));
 
-        if (!driver->is_querying && !driver->is_shutting_up && driver->is_speaking)
-	{	
 #ifdef FESTIVAL_DEBUG_CONTROL
-	    fprintf (stderr, "\nSENT QUERY");
+	fprintf (stderr, "\nSENT QUERY");
 #endif
-	    festival_synthesis_driver_say_raw (driver, "(audio_mode 'query)\n");
-	    driver->is_querying = TRUE;
-	}
+	festival_synthesis_driver_say_raw (driver, "(audio_mode 'query)\n");
     
         return TRUE;
 }
@@ -928,6 +988,7 @@ festival_synthesis_driver_say_ (FestivalSynthesisDriver *d,
 	    festival_synthesis_driver_say_raw (d, s->voice);
 	    speaker_refresh_parameters (SPEAKER(s));
 	    d->last_speaker = s;
+	    g_io_channel_set_encoding(d->channel_sock, encoding_get(d->last_speaker), NULL);
 	}
 
 	clb_list_free (d->crt_clbs);
